"""チャート作成のためのモジュール"""

from __future__ import annotations

from typing import TYPE_CHECKING, Literal

import altair as alt

if TYPE_CHECKING:
    from kabukit.core.prices import Prices


def plot_prices(
    prices: Prices,
    kind: Literal["candlestick"] = "candlestick",
) -> alt.VConcatChart:
    if kind == "candlestick":
        chart_price = plot_prices_candlestick(prices)
        chart_price_volume = plot_prices_volume(prices)
        return alt.vconcat(chart_price, chart_price_volume)

    raise NotImplementedError  # pyright: ignore[reportUnreachable]


def plot_prices_candlestick(prices: Prices) -> alt.LayerChart:
    rule = alt.Chart(prices.data, mark="rule").encode(y="Low:Q", y2="High:Q")
    bar = alt.Chart(prices.data, mark="bar").encode(y="Open:Q", y2="Close:Q")

    color_condition = alt.condition(
        "datum.Open < datum.Close",
        alt.value("#ff3030"),
        alt.value("#3030ff"),
    )

    return alt.layer(rule, bar, height=200).encode(
        x=alt.X("Date:T", axis=alt.Axis(title="日付", format="%Y-%m-%d")),
        y=alt.Y(title="株価", scale=alt.Scale(zero=False)),
        color=color_condition,
        tooltip=[
            alt.Tooltip("Date:T", title="日付"),
            alt.Tooltip("Open:Q", title="始値"),
            alt.Tooltip("High:Q", title="高値"),
            alt.Tooltip("Low:Q", title="安値"),
            alt.Tooltip("Close:Q", title="終値"),
        ],
    )


def plot_prices_volume(prices: Prices) -> alt.Chart:
    return alt.Chart(prices.data, mark="bar", height=50).encode(
        x=alt.X("Date:T", axis=alt.Axis(title="日付", format="%Y-%m-%d")),
        y=alt.Y("Volume:Q", title="出来高"),
        tooltip=[
            alt.Tooltip("Date:T", title="日付"),
            alt.Tooltip("Volume:Q", title="出来高"),
        ],
    )



from __future__ import annotations

import io
import os
import zipfile
from enum import StrEnum
from typing import TYPE_CHECKING

from polars import DataFrame

from kabukit.core.client import Client
from kabukit.utils.config import load_dotenv
from kabukit.utils.params import get_params

from .doc import clean_csv, clean_list, read_csv

if TYPE_CHECKING:
    import datetime

    from httpx import Response
    from httpx._types import QueryParamTypes

API_VERSION = "v2"
BASE_URL = f"https://api.edinet-fsa.go.jp/api/{API_VERSION}"


class AuthKey(StrEnum):
    """Environment variable keys for EDINET authentication."""

    API_KEY = "EDINET_API_KEY"


class EdinetClient(Client):
    def __init__(self, api_key: str | None = None) -> None:
        super().__init__(BASE_URL)
        self.set_api_key(api_key)

    def set_api_key(self, api_key: str | None = None) -> None:
        if api_key is None:
            load_dotenv()
            api_key = os.environ.get(AuthKey.API_KEY)

        if api_key:
            self.client.params = {"Subscription-Key": api_key}

    async def get(self, url: str, params: QueryParamTypes) -> Response:
        resp = await self.client.get(url, params=params)
        resp.raise_for_status()
        return resp

    async def get_count(self, date: str | datetime.date) -> int:
        params = get_params(date=date, type=1)
        resp = await self.get("/documents.json", params)
        data = resp.json()
        metadata = data["metadata"]

        if metadata["status"] != "200":
            return 0

        return metadata["resultset"]["count"]

    async def get_list(self, date: str | datetime.date) -> DataFrame:
        params = get_params(date=date, type=2)
        resp = await self.get("/documents.json", params)
        data = resp.json()

        if "results" not in data:
            return DataFrame()

        df = DataFrame(data["results"], infer_schema_length=None)

        if df.is_empty():
            return df

        return clean_list(df, date)

    async def get_document(self, doc_id: str, doc_type: int) -> Response:
        params = get_params(type=doc_type)
        return await self.get(f"/documents/{doc_id}", params)

    async def get_pdf(self, doc_id: str) -> bytes:
        resp = await self.get_document(doc_id, doc_type=2)
        if resp.headers["content-type"] == "application/pdf":
            return resp.content

        msg = "PDF is not available."
        raise ValueError(msg)

    async def get_zip(self, doc_id: str, doc_type: int) -> bytes:
        resp = await self.get_document(doc_id, doc_type=doc_type)
        if resp.headers["content-type"] == "application/octet-stream":
            return resp.content

        msg = "ZIP is not available."
        raise ValueError(msg)

    async def get_csv(self, doc_id: str) -> DataFrame:
        content = await self.get_zip(doc_id, doc_type=5)
        buffer = io.BytesIO(content)

        with zipfile.ZipFile(buffer) as zf:
            for info in zf.infolist():
                if info.filename.endswith(".csv"):
                    with zf.open(info) as f:
                        df = read_csv(f.read())
                        return clean_csv(df, doc_id)

        msg = "CSV is not available."
        raise ValueError(msg)



from __future__ import annotations

import datetime
from typing import TYPE_CHECKING, Any

if TYPE_CHECKING:
    from collections.abc import Iterator


def iter_items(kwargs: dict[str, Any]) -> Iterator[tuple[str, Any]]:
    for key, value in kwargs.items():
        if value is None:
            continue

        if key == "from_":
            yield "from", value
        else:
            yield key, value


def get_params(**kwargs: Any) -> dict[str, Any]:
    params: dict[str, Any] = {}

    for key, value in iter_items(kwargs):
        if isinstance(value, datetime.date):
            params[key] = date_to_str(value)
        elif isinstance(value, str | int | float | bool):
            params[key] = value
        else:
            params[key] = str(value)

    return params


def date_to_str(date: str | datetime.date) -> str:
    """Convert a date object or string to a YYYY-MM-DD string.

    Args:
        date: The date to convert.

    Returns:
        The date as a YYYY-MM-DD string.
    """
    if isinstance(date, datetime.date):
        return date.strftime("%Y-%m-%d")

    return date



from __future__ import annotations

from typing import TYPE_CHECKING

from kabukit.utils import concurrent
from kabukit.utils.date import get_dates

from .client import EdinetClient

if TYPE_CHECKING:
    import datetime
    from collections.abc import Iterable

    from polars import DataFrame

    from kabukit.utils.concurrent import Callback, Progress


async def fetch(
    resource: str,
    args: Iterable[str | datetime.date],
    /,
    max_concurrency: int | None = None,
    progress: Progress | None = None,
    callback: Callback | None = None,
) -> DataFrame:
    """引数に対応する各種データを取得し、単一のDataFrameにまとめて返す。

    Args:
        resource (str): 取得するデータの種類。EdinetClientのメソッド名から"get_"を
            除いたものを指定する。
        args (Iterable[str | datetime.date]): 取得対象の引数のリスト。
        max_concurrency (int | None, optional): 同時に実行するリクエストの最大数。
            指定しないときはデフォルト値が使用される。
        progress (Progress | None, optional): 進捗表示のための関数。
            tqdm, marimoなどのライブラリを使用できる。
            指定しないときは進捗表示は行われない。
        callback (Callback | None, optional): 各DataFrameに対して適用する
            コールバック関数。指定しないときはそのままのDataFrameが使用される。

    Returns:
        DataFrame:
            すべての銘柄の財務情報を含む単一のDataFrame。
    """
    return await concurrent.fetch(
        EdinetClient,
        resource,
        args,
        max_concurrency=max_concurrency,
        progress=progress,
        callback=callback,
    )


async def fetch_list(
    days: int | None = None,
    years: int | None = None,
    limit: int | None = None,
    max_concurrency: int | None = None,
    progress: Progress | None = None,
    callback: Callback | None = None,
) -> DataFrame:
    """過去days日またはyears年の文書一覧を取得し、単一のDataFrameにまとめて返す。

    Args:
        days (int | None): 過去days日の日付リストを取得する。
        years (int | None): 過去years年の日付リストを取得する。
            daysが指定されている場合は無視される。
        max_concurrency (int | None, optional): 同時に実行するリクエストの最大数。
            指定しないときはデフォルト値が使用される。
        progress (Progress | None, optional): 進捗表示のための関数。
            tqdm, marimoなどのライブラリを使用できる。
            指定しないときは進捗表示は行われない。
        callback (Callback | None, optional): 各DataFrameに対して適用する
            コールバック関数。指定しないときはそのままのDataFrameが使用される。

    Returns:
        DataFrame:
            文書一覧を含む単一のDataFrame。
    """
    dates = get_dates(days=days, years=years)

    if limit is not None:
        dates = dates[:limit]

    df = await fetch(
        "list",
        dates,
        max_concurrency=max_concurrency,
        progress=progress,
        callback=callback,
    )
    return df.sort("Date")


async def fetch_csv(
    doc_ids: Iterable[str],
    /,
    limit: int | None = None,
    max_concurrency: int | None = None,
    progress: Progress | None = None,
    callback: Callback | None = None,
) -> DataFrame:
    """文書をCSV形式で取得し、単一のDataFrameにまとめて返す。

    Args:
        doc_ids (Iterable[str]): 取得対象の文書IDのリスト。
        max_concurrency (int | None, optional): 同時に実行するリクエストの最大数。
            指定しないときはデフォルト値が使用される。
        progress (Progress | None, optional): 進捗表示のための関数。
            tqdm, marimoなどのライブラリを使用できる。
            指定しないときは進捗表示は行われない。
        callback (Callback | None, optional): 各DataFrameに対して適用する
            コールバック関数。指定しないときはそのままのDataFrameが使用される。

    Returns:
        DataFrame:
            文書含む単一のDataFrame。
    """
    doc_ids = list(doc_ids)

    if limit is not None:
        doc_ids = doc_ids[:limit]

    df = await fetch(
        "csv",
        doc_ids,
        max_concurrency=max_concurrency,
        progress=progress,
        callback=callback,
    )
    return df.sort("docID")



from __future__ import annotations

import polars as pl
from polars import DataFrame


def clean(df: DataFrame) -> DataFrame:
    return df.with_columns(
        pl.col("Date").str.to_date("%Y-%m-%d"),
        pl.col("^.*CodeName$", "ScaleCategory").cast(pl.Categorical),
    ).drop("^.+Code$", "CompanyNameEnglish")


async def get_codes() -> list[str]:
    """銘柄コードのリストを返す。

    市場「TOKYO PRO MARKET」と業種「その他」を除外した銘柄を対象とする。
    """
    from .client import JQuantsClient

    async with JQuantsClient() as client:
        info = await client.get_info()

    return (
        info.filter(
            pl.col("MarketCodeName") != "TOKYO PRO MARKET",
            pl.col("Sector17CodeName") != "その他",
        )
        .get_column("Code")
        .to_list()
    )



from __future__ import annotations

from typing import TYPE_CHECKING

import polars as pl

from .base import Base

if TYPE_CHECKING:
    from datetime import timedelta
    from typing import Self

    from polars import Expr

    from .statements import Statements


class Prices(Base):
    def truncate(self, every: str | timedelta | Expr) -> Self:
        data = (
            self.data.group_by(pl.col("Date").dt.truncate(every), "Code")
            .agg(
                pl.col("Open").drop_nulls().first(),
                pl.col("High").max(),
                pl.col("Low").min(),
                pl.col("Close").drop_nulls().last(),
                pl.col("Volume").sum(),
                pl.col("TurnoverValue").sum(),
            )
            .sort("Code", "Date")
        )
        return self.__class__(data)

    def with_adjusted_shares(self, statements: Statements) -> Self:
        """日次の調整済み株式数を計算し、列として追加する。

        決算短信で報告される株式数（例：発行済株式総数）は、四半期ごとなど
        特定の日付のデータです。一方で、株式分割や併合は日々発生し、株式数を
        変動させます。
        このメソッドは、直近の決算で報告された株式数を、日々の調整係数
        (`AdjustmentFactor`) を用いて補正し、日次ベースの時系列データとして
        提供します。これにより、日々の時価総額計算などが正確に行えるようになります。

        具体的には、`statements`から`IssuedShares`（発行済株式総数）と
        `TreasuryShares`（自己株式数）を取得し、それぞれを調整します。
        計算結果は、元の列名との混同を避けるため、接頭辞`Adjusted`を付与した
        新しい列（`AdjustedIssuedShares`, `AdjustedTreasuryShares`）として
        追加されます。

        .. note::
            この計算は、決算発表間の株式数の変動が、株式分割・併合
            （`AdjustmentFactor`）にのみ起因すると仮定しています。
            期中に行われる増資や自己株式取得など、`AdjustmentFactor`に
            反映されないイベントによる株式数の変動は考慮されません。

        Args:
            statements (Statements): 財務データを提供する`Statements`オブジェクト。

        Returns:
            Self: `AdjustedIssuedShares`および`AdjustedTreasuryShares`列が
            追加された、新しいPricesオブジェクト。
        """
        shares = statements.shares().rename({"Date": "ReportDate"})

        adjusted = (
            self.data.join_asof(
                shares,
                left_on="Date",
                right_on="ReportDate",
                by="Code",
                check_sortedness=False,
            )
            .with_columns(
                (1.0 / pl.col("AdjustmentFactor"))
                .cum_prod()
                .over("Code", "ReportDate")
                .alias("CumulativeRatio"),
            )
            .with_columns(
                (pl.col("IssuedShares", "TreasuryShares") * pl.col("CumulativeRatio"))
                .round(0)
                .cast(pl.Int64)
                .name.prefix("Adjusted"),
            )
            .select("Date", "Code", "AdjustedIssuedShares", "AdjustedTreasuryShares")
        )

        data = self.data.join(adjusted, on=["Date", "Code"], how="left")

        return self.__class__(data)

    def with_market_cap(self) -> Self:
        """時価総額を計算し、列として追加する。

        このメソッドは、日々の調整前終値 (`RawClose`) と、調整済みの発行済株式数
        (`AdjustedIssuedShares`) および自己株式数 (`AdjustedTreasuryShares`)
        を基に、日次ベースの時価総額を計算します。

        計算式:
            時価総額 = 調整前終値 * (調整済み発行済株式数 - 調整済み自己株式数)

        Note:
            このメソッドを呼び出す前に、`with_adjusted_shares()` を
            実行して、調整済みの株式数列を事前に計算しておく必要があります。

        Returns:
            Self: `MarketCap` 列が追加された、新しいPricesオブジェクト。
        """
        shares = pl.col("AdjustedIssuedShares") - pl.col("AdjustedTreasuryShares")
        data = self.data.with_columns(
            (pl.col("RawClose") * shares).round(0).alias("MarketCap"),
        )
        return self.__class__(data)

    def with_equity(self, statements: Statements) -> Self:
        """時系列の純資産を列として追加する。

        Args:
            statements (Statements): 財務データを提供する`Statements`オブジェクト。

        Returns:
            Self: `Equity` 列が追加された、新しいPricesオブジェクト。
        """
        data = self.data.join_asof(
            statements.equity(),
            on="Date",
            by="Code",
            check_sortedness=False,
        )
        return self.__class__(data)

    def with_forecast_profit(self, statements: Statements) -> Self:
        """時系列の予想純利益を列として追加する。

        Args:
            statements (Statements): 財務データを提供する`Statements`オブジェクト。

        Returns:
            Self: `ForecastProfit` 列が追加された、新しいPricesオブジェクト。
        """
        data = self.data.join_asof(
            statements.forecast_profit(),
            on="Date",
            by="Code",
            check_sortedness=False,
        )
        return self.__class__(data)

    def with_forecast_dividend(self, statements: Statements) -> Self:
        """時系列の予想年間配当総額を列として追加する。

        Args:
            statements (Statements): 財務データを提供する`Statements`オブジェクト。

        Returns:
            Self: `ForecastDividend` 列が追加された、新しいPricesオブジェクト。
        """
        data = self.data.join_asof(
            statements.forecast_dividend(),
            on="Date",
            by="Code",
            check_sortedness=False,
        )
        return self.__class__(data)



from __future__ import annotations

from typing import TYPE_CHECKING

import polars as pl

from .base import Base

if TYPE_CHECKING:
    from polars import DataFrame


class Statements(Base):
    def shares(self) -> DataFrame:
        """発行済株式数を取得する。"""
        return self.data.filter(
            pl.col("IssuedShares").is_not_null(),
        ).select(
            "Date",
            "Code",
            "IssuedShares",
            "TreasuryShares",
            "AverageOutstandingShares",
        )

    def equity(self) -> DataFrame:
        """Statementsデータから純資産を抽出する。

        Returns:
            DataFrame: Date, Code, Equity を含むDataFrame
        """
        return self.data.filter(
            pl.col("Equity").is_not_null(),
        ).select("Date", "Code", "Equity")

    def forecast_profit(self) -> DataFrame:
        """Statementsデータから予想純利益を抽出する。

        Returns:
            DataFrame: Date, Code, ForecastProfit を含むDataFrame
        """
        return (
            self.data.with_columns(
                pl.when(pl.col("TypeOfDocument").str.starts_with("FY"))
                .then(pl.col("NextYearForecastProfit"))
                .otherwise(pl.col("ForecastProfit"))
                .alias("ForecastProfit"),
            )
            .filter(pl.col("ForecastProfit").is_not_null())
            .select("Date", "Code", "ForecastProfit")
        )

    def forecast_dividend(self) -> DataFrame:
        """予想年間配当総額を抽出する。

        Returns:
            DataFrame: Date, Code, ForecastDividend を含むDataFrame
        """
        # 予想株式数を計算
        forecast_shares = (
            pl.when(pl.col("TypeOfDocument").str.starts_with("FY"))
            .then(
                pl.col("NextYearForecastProfit")
                / pl.col("NextYearForecastEarningsPerShare"),
            )
            .otherwise(pl.col("ForecastProfit") / pl.col("ForecastEarningsPerShare"))
            .alias("ForecastShares")
        )

        # 年間配当総額を計算
        annual_forecast_dividend = (
            pl.when(pl.col("TypeOfDocument").str.starts_with("FY"))
            .then(
                pl.col("NextYearForecastDividendPerShareAnnual")
                * pl.col("ForecastShares"),
            )
            .otherwise(
                pl.col("ForecastDividendPerShareAnnual") * pl.col("ForecastShares"),
            )
            .round(0)
            .alias("ForecastDividend")
        )

        return (
            self.data.with_columns(forecast_shares)
            .with_columns(annual_forecast_dividend)
            .filter(pl.col("ForecastDividend").is_not_null())
            .select("Date", "Code", "ForecastDividend")
        )



from __future__ import annotations

from pathlib import Path

import dotenv
from platformdirs import user_config_dir


def get_dotenv_path() -> Path:
    """Return the path to the .env file in the user config directory."""
    config_dir = Path(user_config_dir("kabukit"))
    config_dir.mkdir(parents=True, exist_ok=True)
    return config_dir / ".env"


def set_key(key: str, value: str) -> tuple[bool | None, str, str]:
    dotenv_path = get_dotenv_path()
    return dotenv.set_key(dotenv_path, key, value)


def load_dotenv() -> bool:
    dotenv_path = get_dotenv_path()
    return dotenv.load_dotenv(dotenv_path)



from __future__ import annotations

from typing import TYPE_CHECKING

import polars as pl

if TYPE_CHECKING:
    from polars import DataFrame


def clean(df: DataFrame) -> DataFrame:
    return df.select(
        pl.col("Date").str.to_date("%Y-%m-%d"),
        "Code",
        Open=pl.col("AdjustmentOpen"),
        High=pl.col("AdjustmentHigh"),
        Low=pl.col("AdjustmentLow"),
        Close=pl.col("AdjustmentClose"),
        UpperLimit=pl.col("UpperLimit").cast(pl.Int8).cast(pl.Boolean),
        LowerLimit=pl.col("LowerLimit").cast(pl.Int8).cast(pl.Boolean),
        Volume=pl.col("AdjustmentVolume"),
        TurnoverValue=pl.col("TurnoverValue"),
        AdjustmentFactor=pl.col("AdjustmentFactor"),
        RawOpen=pl.col("Open"),
        RawHigh=pl.col("High"),
        RawLow=pl.col("Low"),
        RawClose=pl.col("Close"),
        RawVolume=pl.col("Volume"),
    )



from __future__ import annotations

from typing import TYPE_CHECKING

from kabukit.utils import concurrent

from .client import JQuantsClient
from .info import get_codes

if TYPE_CHECKING:
    from collections.abc import Iterable

    from polars import DataFrame

    from kabukit.utils.concurrent import Callback, Progress


async def fetch(
    resource: str,
    codes: Iterable[str],
    /,
    max_concurrency: int | None = None,
    progress: Progress | None = None,
    callback: Callback | None = None,
) -> DataFrame:
    """複数の銘柄の各種データを取得し、単一のDataFrameにまとめて返す。

    Args:
        resource (str): 取得するデータの種類。JQuantsClientのメソッド名から"get_"を
            除いたものを指定する。
        codes (Iterable[str]): 取得対象の銘柄コードのリスト。
        max_concurrency (int | None, optional): 同時に実行するリクエストの最大数。
            指定しないときはデフォルト値が使用される。
        progress (Progress | None, optional): 進捗表示のための関数。
            tqdm, marimoなどのライブラリを使用できる。
            指定しないときは進捗表示は行われない。
        callback (Callback | None, optional): 各DataFrameに対して適用する
            コールバック関数。指定しないときはそのままのDataFrameが使用される。

    Returns:
        DataFrame:
            すべての銘柄の財務情報を含む単一のDataFrame。
    """
    data = await concurrent.fetch(
        JQuantsClient,
        resource,
        codes,
        max_concurrency=max_concurrency,
        progress=progress,
        callback=callback,
    )
    return data.sort("Code", "Date")


async def fetch_all(
    resource: str,
    /,
    limit: int | None = None,
    max_concurrency: int | None = None,
    progress: Progress | None = None,
    callback: Callback | None = None,
) -> DataFrame:
    """全銘柄の各種データを取得し、単一のDataFrameにまとめて返す。

    Args:
        resource (str): 取得するデータの種類。JQuantsClientのメソッド名から"get_"を
            除いたものを指定する。
        limit (int | None, optional): 取得する銘柄数の上限。
            指定しないときはすべての銘柄が対象となる。
        max_concurrency (int | None, optional): 同時に実行するリクエストの最大数。
            指定しないときはデフォルト値が使用される。
        progress (Progress | None, optional): 進捗表示のための関数。
            tqdm, marimoなどのライブラリを使用できる。
            指定しないときは進捗表示は行われない。
        callback (Callback | None, optional): 各DataFrameに対して適用する
            コールバック関数。指定しないときはそのままのDataFrameが使用される。

    Returns:
        DataFrame:
            すべての銘柄の財務情報を含む単一のDataFrame。
    """

    codes = await get_codes()
    codes = codes[:limit]

    return await fetch(
        resource,
        codes,
        max_concurrency=max_concurrency,
        progress=progress,
        callback=callback,
    )



from .core.info import Info
from .core.list import List
from .core.prices import Prices
from .core.reports import Reports
from .core.statements import Statements
from .edinet.client import EdinetClient
from .jquants.client import JQuantsClient

__all__ = [
    "EdinetClient",
    "Info",
    "JQuantsClient",
    "List",
    "Prices",
    "Reports",
    "Statements",
]



from __future__ import annotations

import datetime
from pathlib import Path
from typing import TYPE_CHECKING

import polars as pl
from platformdirs import user_cache_dir

if TYPE_CHECKING:
    from collections.abc import Iterable
    from typing import Any, Self

    from polars import DataFrame
    from polars._typing import IntoExprColumn


class Base:
    data: DataFrame

    def __init__(self, data: DataFrame) -> None:
        self.data = data

    @classmethod
    def data_dir(cls) -> Path:
        clsname = cls.__name__.lower()
        return Path(user_cache_dir("kabukit")) / clsname

    def write(self) -> Path:
        data_dir = self.data_dir()
        data_dir.mkdir(parents=True, exist_ok=True)
        path = datetime.datetime.today().strftime("%Y%m%d")  # noqa: DTZ002
        filename = data_dir / f"{path}.parquet"
        self.data.write_parquet(filename)
        return filename

    @classmethod
    def read(cls, path: str | None = None) -> Self:
        data_dir = cls.data_dir()

        if path:
            filename = data_dir / path
        else:
            filenames = sorted(data_dir.glob("*.parquet"))
            if not filenames:
                msg = f"No data found in {data_dir}"
                raise FileNotFoundError(msg)

            filename = filenames[-1]

        data = pl.read_parquet(filename)
        return cls(data)

    def filter(
        self,
        *predicates: IntoExprColumn | Iterable[IntoExprColumn] | bool | list[bool],
        **constraints: Any,
    ) -> Self:
        """Filter the data with given predicates and constraints."""
        data = self.data.filter(*predicates, **constraints)
        return self.__class__(data)



from __future__ import annotations

from enum import Enum
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from polars import DataFrame


class BaseColumns(Enum):
    @classmethod
    def rename(cls, df: DataFrame, *, strict: bool = False) -> DataFrame:
        """DataFrameの列名を日本語から英語に変換する。"""
        return df.rename({x.name: x.value for x in cls}, strict=strict)


class InfoColumns(BaseColumns):
    Date = "日付"
    Code = "銘柄コード"
    CompanyName = "会社名"
    Sector17CodeName = "17業種コード名"
    Sector33CodeName = "33業種コード名"
    ScaleCategory = "規模コード"
    MarketCodeName = "市場区分名"
    MarginCodeName = "貸借信用区分名"


class PriceColumns(BaseColumns):
    Date = "日付"
    Code = "銘柄コード"
    Open = "始値"
    High = "高値"
    Low = "安値"
    Close = "終値"
    UpperLimit = "ストップ高?"
    LowerLimit = "ストップ安?"
    Volume = "出来高"
    TurnoverValue = "売買代金"
    AdjustmentFactor = "調整係数"
    RawOpen = "調整前始値"
    RawHigh = "調整前高値"
    RawLow = "調整前安値"
    RawClose = "調整前終値"
    RawVolume = "調整前出来高"


class StatementColumns(BaseColumns):
    Date = "日付"
    DisclosedDate = "開示日"
    DisclosedTime = "開示時刻"
    Code = "銘柄コード"
    DisclosureNumber = "開示番号"
    TypeOfDocument = "開示書類種別"
    TypeOfCurrentPeriod = "当会計期間の種類"
    CurrentPeriodStartDate = "当会計期間開始日"
    CurrentPeriodEndDate = "当会計期間終了日"
    CurrentFiscalYearStartDate = "当事業年度開始日"
    CurrentFiscalYearEndDate = "当事業年度終了日"
    NextFiscalYearStartDate = "翌事業年度開始日"
    NextFiscalYearEndDate = "翌事業年度終了日"

    NetSales = "売上高"
    OperatingProfit = "営業利益"
    OrdinaryProfit = "経常利益"
    Profit = "当期純利益"
    EarningsPerShare = "一株あたり当期純利益"
    DilutedEarningsPerShare = "潜在株式調整後一株あたり当期純利益"

    TotalAssets = "総資産"
    Equity = "純資産"
    EquityToAssetRatio = "自己資本比率"
    BookValuePerShare = "一株あたり純資産"

    CashFlowsFromOperatingActivities = "営業活動によるキャッシュフロー"
    CashFlowsFromInvestingActivities = "投資活動によるキャッシュフロー"
    CashFlowsFromFinancingActivities = "財務活動によるキャッシュフロー"
    CashAndEquivalents = "現金及び現金同等物期末残高"

    ResultDividendPerShare1stQuarter = "一株あたり配当実績_第1四半期末"
    ResultDividendPerShare2ndQuarter = "一株あたり配当実績_第2四半期末"
    ResultDividendPerShare3rdQuarter = "一株あたり配当実績_第3四半期末"
    ResultDividendPerShareFiscalYearEnd = "一株あたり配当実績_期末"
    ResultDividendPerShareAnnual = "一株あたり配当実績_合計"
    ResultTotalDividendPaidAnnual = "配当金総額"
    ResultPayoutRatioAnnual = "配当性向"

    ForecastDividendPerShare1stQuarter = "一株あたり配当予想_第1四半期末"
    ForecastDividendPerShare2ndQuarter = "一株あたり配当予想_第2四半期末"
    ForecastDividendPerShare3rdQuarter = "一株あたり配当予想_第3四半期末"
    ForecastDividendPerShareFiscalYearEnd = "一株あたり配当予想_期末"
    ForecastDividendPerShareAnnual = "一株あたり配当予想_合計"
    ForecastTotalDividendPaidAnnual = "予想配当金総額"
    ForecastPayoutRatioAnnual = "予想配当性向"

    NextYearForecastDividendPerShare1stQuarter = "一株あたり配当予想_翌事業年度第1四半期末"
    NextYearForecastDividendPerShare2ndQuarter = "一株あたり配当予想_翌事業年度第2四半期末"
    NextYearForecastDividendPerShare3rdQuarter = "一株あたり配当予想_翌事業年度第3四半期末"
    NextYearForecastDividendPerShareFiscalYearEnd = "一株あたり配当予想_翌事業年度期末"
    NextYearForecastDividendPerShareAnnual = "一株あたり配当予想_翌事業年度合計"
    NextYearForecastPayoutRatioAnnual = "翌事業年度予想配当性向"

    ForecastNetSales2ndQuarter = "売上高_予想_第2四半期末"
    ForecastOperatingProfit2ndQuarter = "営業利益_予想_第2四半期末"
    ForecastOrdinaryProfit2ndQuarter = "経常利益_予想_第2四半期末"
    ForecastProfit2ndQuarter = "当期純利益_予想_第2四半期末"
    ForecastEarningsPerShare2ndQuarter = "一株あたり当期純利益_予想_第2四半期末"

    NextYearForecastNetSales2ndQuarter = "売上高_予想_翌事業年度第2四半期末"
    NextYearForecastOperatingProfit2ndQuarter = "営業利益_予想_翌事業年度第2四半期末"
    NextYearForecastOrdinaryProfit2ndQuarter = "経常利益_予想_翌事業年度第2四半期末"
    NextYearForecastProfit2ndQuarter = "当期純利益_予想_翌事業年度第2四半期末"
    NextYearForecastEarningsPerShare2ndQuarter = "一株あたり当期純利益_予想_翌事業年度第2四半期末"

    ForecastNetSales = "売上高_予想_期末"
    ForecastOperatingProfit = "営業利益_予想_期末"
    ForecastOrdinaryProfit = "経常利益_予想_期末"
    ForecastProfit = "当期純利益_予想_期末"
    ForecastEarningsPerShare = "一株あたり当期純利益_予想_期末"

    NextYearForecastNetSales = "売上高_予想_翌事業年度期末"
    NextYearForecastOperatingProfit = "営業利益_予想_翌事業年度期末"
    NextYearForecastOrdinaryProfit = "経常利益_予想_翌事業年度期末"
    NextYearForecastProfit = "当期純利益_予想_翌事業年度期末"
    NextYearForecastEarningsPerShare = "一株あたり当期純利益_予想_翌事業年度期末"

    MaterialChangesInSubsidiaries = "期中における重要な子会社の異動"
    SignificantChangesInTheScopeOfConsolidation = "期中における連結範囲の重要な変更"
    ChangesBasedOnRevisionsOfAccountingStandard = "会計基準等の改正に伴う会計方針の変更"
    ChangesOtherThanOnesBasedOnRevisionsOfAccountingStandard = "会計基準等の改正に伴う変更以外の会計方針の変更"
    ChangesInAccountingEstimates = "会計上の見積りの変更"
    RetrospectiveRestatement = "修正再表示"

    IssuedShares = "期末発行済株式数" # 自己株式を含む (NumberOfIssuedAndOutstandingSharesAtTheEndOfFiscalYearIncludingTreasuryStock)
    TreasuryShares = "期末自己株式数" # (NumberOfTreasuryStockAtTheEndOfFiscalYear)
    AverageOutstandingShares = "期中平均株式数" # 自己株式を除く。EPSなどの計算に使用される (AverageNumberOfShares)

    NonConsolidatedNetSales = "売上高_非連結"
    NonConsolidatedOperatingProfit = "営業利益_非連結"
    NonConsolidatedOrdinaryProfit = "経常利益_非連結"
    NonConsolidatedProfit = "当期純利益_非連結"
    NonConsolidatedEarningsPerShare = "一株あたり当期純利益_非連結"

    NonConsolidatedTotalAssets = "総資産_非連結"
    NonConsolidatedEquity = "純資産_非連結"
    NonConsolidatedEquityToAssetRatio = "自己資本比率_非連結"
    NonConsolidatedBookValuePerShare = "一株あたり純資産_非連結"

    ForecastNonConsolidatedNetSales2ndQuarter = "売上高_予想_第2四半期末_非連結"
    ForecastNonConsolidatedOperatingProfit2ndQuarter = "営業利益_予想_第2四半期末_非連結"
    ForecastNonConsolidatedOrdinaryProfit2ndQuarter = "経常利益_予想_第2四半期末_非連結"
    ForecastNonConsolidatedProfit2ndQuarter = "当期純利益_予想_第2四半期末_非連結"
    ForecastNonConsolidatedEarningsPerShare2ndQuarter = "一株あたり当期純利益_予想_第2四半期末_非連結"

    NextYearForecastNonConsolidatedNetSales2ndQuarter = "売上高_予想_翌事業年度第2四半期末_非連結"
    NextYearForecastNonConsolidatedOperatingProfit2ndQuarter = "営業利益_予想_翌事業年度第2四半期末_非連結"
    NextYearForecastNonConsolidatedOrdinaryProfit2ndQuarter = "経常利益_予想_翌事業年度第2四半期末_非連結"
    NextYearForecastNonConsolidatedProfit2ndQuarter = "当期純利益_予想_翌事業年度第2四半期末_非連結"
    NextYearForecastNonConsolidatedEarningsPerShare2ndQuarter = "一株あたり当期純利益_予想_翌事業年度第2四半期末_非連結"

    ForecastNonConsolidatedNetSales = "売上高_予想_期末_非連結"
    ForecastNonConsolidatedOperatingProfit = "営業利益_予想_期末_非連結"
    ForecastNonConsolidatedOrdinaryProfit = "経常利益_予想_期末_非連結"
    ForecastNonConsolidatedProfit = "当期純利益_予想_期末_非連結"
    ForecastNonConsolidatedEarningsPerShare = "一株あたり当期純利益_予想_期末_非連結"

    NextYearForecastNonConsolidatedNetSales = "売上高_予想_翌事業年度期末_非連結"
    NextYearForecastNonConsolidatedOperatingProfit = "営業利益_予想_翌事業年度期末_非連結"
    NextYearForecastNonConsolidatedOrdinaryProfit = "経常利益_予想_翌事業年度期末_非連結"
    NextYearForecastNonConsolidatedProfit = "当期純利益_予想_翌事業年度期末_非連結"
    NextYearForecastNonConsolidatedEarningsPerShare = "一株あたり当期純利益_予想_翌事業年度期末_非連結"


def rename(df: DataFrame, *, strict: bool = False) -> DataFrame:
    """DataFrameの列名を日本語から英語に変換する。"""
    for enum in (InfoColumns, PriceColumns, StatementColumns):
        df = enum.rename(df, strict=strict)
    return df



from __future__ import annotations

from typing import Annotated

import typer
from async_typer import AsyncTyper  # pyright: ignore[reportMissingTypeStubs]
from httpx import HTTPStatusError
from typer import Exit, Option

app = AsyncTyper(
    add_completion=False,
    help="J-QuantsまたはEDINETの認証トークンを保存します。",
)


async def auth_jquants(mailaddress: str, password: str) -> None:
    """J-Quants APIの認証を行い、トークンを設定ファイルに保存します。"""
    from kabukit.jquants.client import JQuantsClient

    async with JQuantsClient() as client:
        try:
            await client.auth(mailaddress, password, save=True)
        except HTTPStatusError:
            typer.echo("認証に失敗しました。")
            raise Exit(1) from None

    typer.echo("J-Quantsのリフレッシュトークン・IDトークンを保存しました。")


Mailaddress = Annotated[
    str,
    Option(prompt=True, help="J-Quantsに登録したメールアドレス。"),
]
Password = Annotated[
    str,
    Option(prompt=True, hide_input=True, help="J-Quantsのパスワード。"),
]


@app.async_command()  # pyright: ignore[reportUnknownMemberType]
async def jquants(mailaddress: Mailaddress, password: Password) -> None:
    """J-Quants APIの認証を行い、トークンを設定ファイルに保存します。(エイリアス: j)"""
    await auth_jquants(mailaddress, password)


@app.async_command(name="j", hidden=True)  # pyright: ignore[reportUnknownMemberType]
async def jquants_alias(mailaddress: Mailaddress, password: Password) -> None:
    await auth_jquants(mailaddress, password)


def auth_edinet(api_key: str) -> None:
    """EDINET APIのAPIキーを設定ファイルに保存します。"""
    from kabukit.utils.config import set_key

    set_key("EDINET_API_KEY", api_key)
    typer.echo("EDINETのAPIキーを保存しました。")


ApiKey = Annotated[str, Option(prompt=True, help="取得したEDINET APIキー。")]


@app.command()
def edinet(api_key: ApiKey) -> None:
    """EDINET APIのAPIキーを設定ファイルに保存します。(エイリアス: e)"""
    auth_edinet(api_key)


@app.command(name="e", hidden=True)
def edinet_alias(api_key: ApiKey) -> None:
    auth_edinet(api_key)


@app.command()
def show() -> None:
    """設定ファイルに保存したトークン・APIキーを表示します。"""
    from dotenv import dotenv_values

    from kabukit.utils.config import get_dotenv_path

    path = get_dotenv_path()
    typer.echo(f"Configuration file: {path}")

    if path.exists():
        config = dotenv_values(path)
        for key, value in config.items():
            typer.echo(f"{key}: {value}")



from __future__ import annotations

import datetime
from typing import TYPE_CHECKING

import polars as pl

if TYPE_CHECKING:
    from polars import DataFrame


def clean_list(df: DataFrame, date: str | datetime.date) -> DataFrame:
    if isinstance(date, str):
        date = datetime.datetime.strptime(date, "%Y-%m-%d").date()  # noqa: DTZ007

    null_columns = [c for c in df.columns if df[c].dtype == pl.Null]

    return (
        df.with_columns(
            pl.col(null_columns).cast(pl.String),
        )
        .with_columns(
            pl.lit(date).alias("Date"),
            pl.col("^.+DateTime$").str.to_datetime("%Y-%m-%d %H:%M", strict=False),
            pl.col("^period.+$").str.to_date("%Y-%m-%d", strict=False),
            pl.col("^.+Flag$").cast(pl.Int8).cast(pl.Boolean),
            pl.col("^.+Code$").cast(pl.String),
        )
        .select("Date", pl.exclude("Date"))
    )


def read_csv(data: bytes) -> DataFrame:
    return pl.read_csv(
        data,
        separator="\t",
        encoding="utf-16-le",
        infer_schema_length=None,
    )


def clean_csv(df: DataFrame, doc_id: str) -> DataFrame:
    return df.select(
        pl.lit(doc_id).alias("docID"),
        pl.all(),
    )



from __future__ import annotations

import asyncio
from dataclasses import dataclass
from typing import TYPE_CHECKING, Any, Protocol

import polars as pl

if TYPE_CHECKING:
    from collections.abc import (
        AsyncIterable,
        AsyncIterator,
        Awaitable,
        Callable,
        Iterable,
    )
    from typing import Any

    from marimo._plugins.stateless.status import progress_bar
    from polars import DataFrame
    from tqdm.asyncio import tqdm

    from kabukit.core.client import Client

    class _Progress(Protocol):
        def __call__(
            self,
            aiterable: AsyncIterable[Any],
            /,
            total: int | None = None,
            *args: Any,
            **kwargs: Any,
        ) -> AsyncIterator[Any]: ...


MAX_CONCURRENCY = 12


async def collect[R](
    awaitables: Iterable[Awaitable[R]],
    /,
    max_concurrency: int | None = None,
) -> AsyncIterator[R]:
    max_concurrency = max_concurrency or MAX_CONCURRENCY
    semaphore = asyncio.Semaphore(max_concurrency)

    async def run(awaitable: Awaitable[R]) -> R:
        async with semaphore:
            return await awaitable

    futures = (run(awaitable) for awaitable in awaitables)

    async for future in asyncio.as_completed(futures):
        yield await future


async def collect_fn[T, R](
    function: Callable[[T], Awaitable[R]],
    args: Iterable[T],
    /,
    max_concurrency: int | None = None,
) -> AsyncIterator[R]:
    max_concurrency = max_concurrency or MAX_CONCURRENCY
    awaitables = (function(arg) for arg in args)

    async for item in collect(awaitables, max_concurrency=max_concurrency):
        yield item


async def concat(
    awaitables: Iterable[Awaitable[DataFrame]],
    /,
    max_concurrency: int | None = None,
) -> DataFrame:
    dfs = collect(awaitables, max_concurrency=max_concurrency)
    dfs = [df async for df in dfs]
    return pl.concat(df for df in dfs if not df.is_empty())


async def concat_fn[T](
    function: Callable[[T], Awaitable[DataFrame]],
    args: Iterable[T],
    /,
    max_concurrency: int | None = None,
) -> DataFrame:
    dfs = collect_fn(function, args, max_concurrency=max_concurrency)
    dfs = [df async for df in dfs]
    return pl.concat(df for df in dfs if not df.is_empty())


type Callback = Callable[[DataFrame], DataFrame | None]
type Progress = type[progress_bar[Any] | tqdm[Any]] | _Progress


@dataclass
class Stream:
    cls: type[Client]
    resource: str
    args: list[Any]
    max_concurrency: int | None = None

    async def __aiter__(self) -> AsyncIterator[DataFrame]:
        async with self.cls() as client:
            fn = getattr(client, f"get_{self.resource}")

            async for df in collect_fn(fn, self.args, self.max_concurrency):
                yield df


async def fetch(
    cls: type[Client],
    resource: str,
    args: Iterable[Any],
    /,
    max_concurrency: int | None = None,
    progress: Progress | None = None,
    callback: Callback | None = None,
) -> DataFrame:
    """各種データを取得し、単一のDataFrameにまとめて返す。

    Args:
        cls (type[Client]): 使用するClientクラス。
            JQuantsClientやEdinetClientなど、Clientを継承したクラス
        resource (str): 取得するデータの種類。Clientのメソッド名から"get_"を
            除いたものを指定する。
        args (Iterable[Any]): 取得対象の引数のリスト。
        max_concurrency (int | None, optional): 同時に実行するリクエストの最大数。
            指定しないときはデフォルト値が使用される。
        progress (Progress | None, optional): 進捗表示のための関数。
            tqdm, marimoなどのライブラリを使用できる。
            指定しないときは進捗表示は行われない。
        callback (Callback | None, optional): 各DataFrameに対して適用する
            コールバック関数。指定しないときはそのままのDataFrameが使用される。

    Returns:
        DataFrame:
            すべての情報を含む単一のDataFrame。
    """
    args = list(args)
    stream = Stream(cls, resource, args, max_concurrency)

    if progress:
        stream = progress(aiter(stream), total=len(args))

    if callback:
        stream = (x if (r := callback(x)) is None else r async for x in stream)

    dfs = [df async for df in stream if not df.is_empty()]
    return pl.concat(dfs) if dfs else pl.DataFrame()



from __future__ import annotations

import datetime


def get_dates(days: int | None = None, years: int | None = None) -> list[datetime.date]:
    """過去days日またはyears年の日付リストを返す。

    Args:
        days (int | None): 過去days日の日付リストを取得する。
        years (int | None): 過去years年の日付リストを取得する。
            daysが指定されている場合は無視される。
    """
    end_date = datetime.date.today()  # noqa: DTZ011

    if days is not None:
        start_date = end_date - datetime.timedelta(days=days)
    elif years is not None:
        start_date = end_date.replace(year=end_date.year - years)
    else:
        msg = "daysまたはyearsのいずれかを指定してください。"
        raise ValueError(msg)

    return [
        start_date + datetime.timedelta(days=i)
        for i in range(1, (end_date - start_date).days + 1)
    ]



"""kabukit CLI."""

from __future__ import annotations

import typer
from async_typer import AsyncTyper  # pyright: ignore[reportMissingTypeStubs]

from . import auth, get

app = AsyncTyper(
    add_completion=False,
    help="J-Quants/EDINETデータツール",
)
app.add_typer(auth.app, name="auth")
app.add_typer(get.app, name="get")


@app.command()
def version() -> None:
    """バージョン情報を表示します。"""
    from importlib.metadata import version

    typer.echo(f"kabukit version: {version('kabukit')}")



from __future__ import annotations

import datetime
import os
from enum import StrEnum
from typing import TYPE_CHECKING

import polars as pl
from polars import DataFrame

from kabukit.core.client import Client
from kabukit.utils.config import load_dotenv, set_key
from kabukit.utils.params import get_params

from . import info, prices, statements

if TYPE_CHECKING:
    from collections.abc import AsyncIterator
    from typing import Any, Self

    from httpx import HTTPStatusError  # noqa: F401
    from httpx._types import QueryParamTypes

API_VERSION = "v1"
BASE_URL = f"https://api.jquants.com/{API_VERSION}"


class AuthKey(StrEnum):
    """J-Quants認証のための環境変数キー。"""

    REFRESH_TOKEN = "JQUANTS_REFRESH_TOKEN"  # noqa: S105
    ID_TOKEN = "JQUANTS_ID_TOKEN"  # noqa: S105


class JQuantsClient(Client):
    """J-Quants APIと対話するためのクライアント。

    API認証トークン（リフレッシュトークンおよびIDトークン）を管理し、
    各種J-Quants APIエンドポイントへアクセスするメソッドを提供する。
    トークンは設定ファイルから読み込まれ、またファイルに保存される。

    Attributes:
        client: APIリクエストを行うための `AsyncClient` インスタンス。
    """

    def __init__(self, id_token: str | None = None) -> None:
        super().__init__(BASE_URL)
        self.set_id_token(id_token)

    def set_id_token(self, id_token: str | None = None) -> None:
        """IDトークンをヘッダーに設定する。

        Args:
            id_token (str | None, optional): 設定するIDトークン。
                Noneの場合、環境変数から読み込む。
        """

        if id_token is None:
            load_dotenv()
            id_token = os.environ.get(AuthKey.ID_TOKEN)

        if id_token:
            self.client.headers["Authorization"] = f"Bearer {id_token}"

    async def post(self, url: str, json: Any | None = None) -> Any:
        """指定されたURLにPOSTリクエストを送信する。

        Args:
            url: POSTリクエストのURLパス。
            json: リクエストボディのJSONペイロード。

        Returns:
            APIからのJSONレスポンス。

        Raises:
            HTTPStatusError: APIリクエストが失敗した場合。
        """
        resp = await self.client.post(url, json=json)
        resp.raise_for_status()
        return resp.json()

    async def get(self, url: str, params: QueryParamTypes | None = None) -> Any:
        """指定されたURLにGETリクエストを送信する。

        Args:
            url (str): GETリクエストのURLパス。
            params (QueryParamTypes | None, optional): リクエストのクエリパラメータ。

        Returns:
            APIからのJSONレスポンス。

        Raises:
            HTTPStatusError: APIリクエストが失敗した場合。
        """
        resp = await self.client.get(url, params=params)
        resp.raise_for_status()
        return resp.json()

    async def auth(
        self,
        mailaddress: str,
        password: str,
        *,
        save: bool = False,
    ) -> Self:
        """認証を行い、トークンを保存する。

        Args:
            mailaddress (str): J-Quantsに登録したメールアドレス。
            password (str): J-Quantsのパスワード。
            save (bool, optional): トークンを環境変数に保存するかどうか。

        Raises:
            HTTPStatusError: APIリクエストが失敗した場合。
        """
        json_data = {"mailaddress": mailaddress, "password": password}
        data = await self.post("/token/auth_user", json=json_data)
        refresh_token = data["refreshToken"]

        url = f"/token/auth_refresh?refreshtoken={refresh_token}"
        data = await self.post(url)
        id_token = data["idToken"]

        if save:
            set_key(AuthKey.REFRESH_TOKEN, refresh_token)
            set_key(AuthKey.ID_TOKEN, id_token)

        self.set_id_token(id_token)
        return self

    async def get_info(
        self,
        code: str | None = None,
        date: str | datetime.date | None = None,
        *,
        clean: bool = True,
    ) -> DataFrame:
        """銘柄情報を取得する。

        Args:
            code (str, optional): 情報を取得する銘柄のコード。
            date (str | datetime.date, optional): 情報を取得する日付。
            clean (bool, optional): 取得したデータをクリーンアップするかどうか。

        Returns:
            銘柄情報を含むDataFrame。

        Raises:
            HTTPStatusError: APIリクエストが失敗した場合。
        """
        params = get_params(code=code, date=date)
        url = "/listed/info"
        data = await self.get(url, params)
        df = DataFrame(data["info"])
        return info.clean(df) if clean else df

    async def iter_pages(
        self,
        url: str,
        params: dict[str, Any] | None,
        name: str,
    ) -> AsyncIterator[DataFrame]:
        """ページ分割されたAPIレスポンスを反復処理する。

        Args:
            url (str): APIエンドポイントのベースURL。
            params (dict[str, Any]): クエリパラメータの辞書。
            name (str): アイテムのリストを含むJSONレスポンスのキー。

        Yields:
            データの各ページに対応するDataFrame。

        Raises:
            HTTPStatusError: APIリクエストが失敗した場合。
        """
        params = params or {}

        while True:
            data = await self.get(url, params)
            yield DataFrame(data[name])
            if "pagination_key" in data:
                params["pagination_key"] = data["pagination_key"]
            else:
                break

    async def get_prices(
        self,
        code: str | None = None,
        date: str | datetime.date | None = None,
        from_: str | datetime.date | None = None,
        to: str | datetime.date | None = None,
        *,
        clean: bool = True,
    ) -> DataFrame:
        """日々の株価四本値を取得する。

        株価は分割・併合を考慮した調整済み株価（小数点第２位四捨五入）と調整前の株価を取得できる。

        Args:
            code (str, optional): 株価を取得する銘柄のコード。
            date (str | datetime.date, optional): 株価を取得する日付。
                `from_`または`to`とは併用不可。
            from_ (str | datetime.date, optional): 取得期間の開始日。
                `date`とは併用不可。
            to (str | datetime.date, optional): 取得期間の終了日。
                `date`とは併用不可。
            clean (bool, optional): 取得したデータをクリーンアップするかどうか。

        Returns:
            日々の株価四本値を含むDataFrame。

        Raises:
            ValueError: `date`と`from_`/`to`の両方が指定された場合。
            HTTPStatusError: APIリクエストが失敗した場合。
        """
        if not date and not code:
            return await self.get_latest_available_prices(clean=clean)

        if date and (from_ or to):
            msg = "dateとfrom/toの両方を指定することはできません。"
            raise ValueError(msg)

        params = get_params(code=code, date=date, from_=from_, to=to)

        url = "/prices/daily_quotes"
        name = "daily_quotes"

        dfs = [df async for df in self.iter_pages(url, params, name)]
        df = pl.concat(dfs)

        if df.is_empty():
            return df

        return prices.clean(df) if clean else df

    async def get_latest_available_prices(
        self,
        num_days: int = 30,
        *,
        clean: bool = True,
    ) -> DataFrame:
        """直近利用可能な日付の株価を取得する。"""
        today = datetime.date.today()  # noqa: DTZ011

        for days in range(num_days):
            date = today - datetime.timedelta(days)
            df = await self.get_prices(date=date, clean=clean)

            if not df.is_empty():
                return df

        return DataFrame()

    async def get_statements(
        self,
        code: str | None = None,
        date: str | datetime.date | None = None,
        *,
        clean: bool = True,
        with_date: bool = True,
    ) -> DataFrame:
        """四半期毎の決算短信サマリーおよび業績・配当の修正に関する開示情報を取得する。

        Args:
            code (str, optional): 財務情報を取得する銘柄のコード。
            date (str | datetime.date, optional): 財務情報を取得する日付。
            clean (bool, optional): 取得したデータをクリーンアップするかどうか。
            with_date (bool, optional): クリーンアップ後に営業日ベースで開示日の翌日を
                計算して`Date`列を追加するかどうか。

        Returns:
            財務情報を含むDataFrame。

        Raises:
            ValueError: `code`と`date`が両方とも指定されない場合。
            HTTPStatusError: APIリクエストが失敗した場合。
        """
        if not code and not date:
            msg = "codeまたはdateのどちらかを指定する必要があります。"
            raise ValueError(msg)

        params = get_params(code=code, date=date)
        url = "/fins/statements"
        name = "statements"

        dfs = [df async for df in self.iter_pages(url, params, name)]
        df = pl.concat(dfs)

        if df.is_empty() or not clean:
            return df

        df = statements.clean(df)

        return statements.with_date(df) if with_date else df

    async def get_announcement(self) -> DataFrame:
        """翌日発表予定の決算情報を取得する。

        Returns:
            開示情報を含むPolars DataFrame。

        Raises:
            HTTPStatusError: APIリクエストが失敗した場合。
        """
        url = "fins/announcement"
        name = "announcement"

        dfs = [df async for df in self.iter_pages(url, {}, name)]
        df = pl.concat(dfs)
        if df.is_empty():
            return df

        return df.with_columns(pl.col("Date").str.to_date("%Y-%m-%d", strict=False))

    async def get_trades_spec(
        self,
        section: str | None = None,
        from_: str | datetime.date | None = None,
        to: str | datetime.date | None = None,
    ) -> DataFrame:
        """投資部門別の情報を取得する。

        Args:
            section: 絞り込み対象のセクション。
            from_: 取得期間の開始日。
            to: 取得期間の終了日。

        Returns:
            投資部門別の情報を含むPolars DataFrame。

        Raises:
            HTTPStatusError: APIリクエストが失敗した場合。
        """
        params = get_params(section=section, from_=from_, to=to)

        url = "/markets/trades_spec"
        name = "trades_spec"

        dfs = [df async for df in self.iter_pages(url, params, name)]
        df = pl.concat(dfs)
        if df.is_empty():
            return df

        return df.with_columns(pl.col("^.*Date$").str.to_date("%Y-%m-%d", strict=False))



from __future__ import annotations

from typing import TYPE_CHECKING

from httpx import AsyncClient

if TYPE_CHECKING:
    from typing import Self


class Client:
    client: AsyncClient

    def __init__(self, base_url: str = "") -> None:
        self.client = AsyncClient(base_url=base_url, timeout=20)

    async def aclose(self) -> None:
        """HTTPクライアントを閉じる。"""
        await self.client.aclose()

    async def __aenter__(self) -> Self:
        return self

    async def __aexit__(self, exc_type, exc_val, exc_tb) -> None:  # pyright: ignore[reportMissingParameterType, reportUnknownParameterType]  # noqa: ANN001
        await self.aclose()



from __future__ import annotations

from typing import TYPE_CHECKING, Annotated, Any

import typer
from async_typer import AsyncTyper  # pyright: ignore[reportMissingTypeStubs]
from typer import Argument

if TYPE_CHECKING:
    from kabukit.core.base import Base

app = AsyncTyper(
    add_completion=False,
    help="J-Quantsからデータを取得します。",
)

Code = Annotated[
    str | None,
    Argument(help="銘柄コード。指定しない場合は全銘柄の情報を取得します。"),
]


@app.async_command()  # pyright: ignore[reportUnknownMemberType]
async def info(code: Code = None) -> None:
    """上場銘柄一覧を取得します。"""
    from kabukit.core.info import Info
    from kabukit.jquants.client import JQuantsClient

    async with JQuantsClient() as client:
        df = await client.get_info(code)

    typer.echo(df)

    if code is None:
        path = Info(df).write()
        typer.echo(f"全銘柄の情報を '{path}' に保存しました。")


async def _fetch(
    code: str | None,
    target: str,
    cls: type[Base],
    fetch_func_name: str,
    message: str,
    **kwargs: Any,
) -> None:
    """財務情報・株価情報を取得するための共通処理"""
    from kabukit.jquants.client import JQuantsClient

    if code is not None:
        async with JQuantsClient() as client:
            df = await getattr(client, fetch_func_name)(code)
        typer.echo(df)
        return

    import tqdm.asyncio

    from kabukit.jquants.concurrent import fetch_all

    df = await fetch_all(target, progress=tqdm.asyncio.tqdm, **kwargs)
    typer.echo(df)
    path = cls(df).write()
    typer.echo(f"全銘柄の{message}を '{path}' に保存しました。")


@app.async_command()  # pyright: ignore[reportUnknownMemberType]
async def statements(code: Code = None) -> None:
    """財務情報を取得します。"""
    from kabukit.core.statements import Statements

    await _fetch(
        code=code,
        target="statements",
        cls=Statements,
        fetch_func_name="get_statements",
        message="財務情報",
    )


@app.async_command()  # pyright: ignore[reportUnknownMemberType]
async def prices(code: Code = None) -> None:
    """株価を取得します。"""
    from kabukit.core.prices import Prices

    await _fetch(
        code=code,
        target="prices",
        cls=Prices,
        fetch_func_name="get_prices",
        message="株価情報",
        max_concurrency=8,
    )


@app.async_command(name="list")  # pyright: ignore[reportUnknownMemberType]
async def list_() -> None:
    """報告書一覧を取得します。"""
    import tqdm.asyncio

    from kabukit.core.list import List
    from kabukit.edinet.concurrent import fetch_list

    df = await fetch_list(years=10, progress=tqdm.asyncio.tqdm)
    typer.echo(df)
    path = List(df).write()
    typer.echo(f"報告書一覧を '{path}' に保存しました。")


@app.async_command()  # pyright: ignore[reportUnknownMemberType]
async def reports() -> None:
    """報告書を取得します。"""
    import polars as pl
    import tqdm.asyncio

    from kabukit.core.list import List
    from kabukit.core.reports import Reports
    from kabukit.edinet.concurrent import fetch_csv

    try:
        df = List.read().data
    except FileNotFoundError:
        await list_()
        df = List.read().data

    lst = df.filter(pl.col("csvFlag"), pl.col("secCode").is_not_null())
    doc_ids = lst["docID"].unique()

    df = await fetch_csv(doc_ids, limit=1000, progress=tqdm.asyncio.tqdm)
    typer.echo(df)
    path = Reports(df).write()
    typer.echo(f"報告書を '{path}' に保存しました。")


@app.async_command(name="all")  # pyright: ignore[reportUnknownMemberType]
async def all_(code: Code = None) -> None:
    """上場銘柄一覧、財務情報、株価、報告書を連続して取得します。"""
    typer.echo("上場銘柄一覧を取得します。")
    await info(code)

    typer.echo("---")
    typer.echo("財務情報を取得します。")
    await statements(code)

    typer.echo("---")
    typer.echo("株価を取得します。")
    await prices(code)

    if code is None:
        typer.echo("---")
        typer.echo("報告書一覧を取得します。")
        await list_()
        typer.echo("---")
        typer.echo("報告書を取得します。")
        await reports()



from __future__ import annotations

import datetime
from functools import cache
from typing import TYPE_CHECKING

import holidays
import polars as pl

if TYPE_CHECKING:
    from polars import DataFrame


def clean(df: DataFrame) -> DataFrame:
    df = df.select(pl.exclude(r"^.*\(REIT\)$"))
    return df.pipe(_rename).pipe(_cast)


def _rename(df: DataFrame) -> DataFrame:
    return df.rename(
        {
            "LocalCode": "Code",
            "NumberOfIssuedAndOutstandingSharesAtTheEndOfFiscalYearIncludingTreasuryStock": "IssuedShares",  # noqa: E501
            "NumberOfTreasuryStockAtTheEndOfFiscalYear": "TreasuryShares",
            "AverageNumberOfShares": "AverageOutstandingShares",
        },
    )


def _cast(df: DataFrame) -> DataFrame:
    return (
        df.with_columns(
            pl.col("^.*Date$").str.to_date("%Y-%m-%d", strict=False),
            pl.col("DisclosedTime").str.to_time("%H:%M:%S", strict=False),
            pl.col("TypeOfCurrentPeriod").cast(pl.Categorical),
        )
        .pipe(_cast_float)
        .pipe(_cast_bool)
    )


def _cast_float(df: DataFrame) -> DataFrame:
    return df.with_columns(
        pl.col(f"^.*{name}.*$").cast(pl.Float64, strict=False)
        for name in [
            "Assets",
            "BookValue",
            "Cash",
            "Distributions",
            "Dividend",
            "Earnings",
            "Equity",
            "NetSales",
            "PayoutRatio",
            "Profit",
        ]
    ).with_columns(
        pl.col(
            "IssuedShares",
            "TreasuryShares",
        ).cast(pl.Int64, strict=False),
        pl.col(
            "AverageOutstandingShares",
        ).cast(pl.Float64, strict=False),
    )


def _cast_bool(df: DataFrame) -> DataFrame:
    columns = df.select(pl.col("^.*Changes.*$")).columns
    columns.append("RetrospectiveRestatement")

    return df.with_columns(
        pl.when(pl.col(col) == "true")
        .then(True)  # noqa: FBT003
        .when(pl.col(col) == "false")
        .then(False)  # noqa: FBT003
        .otherwise(None)
        .alias(col)
        for col in columns
    )


@cache
def get_holidays(year: int | None = None, n: int = 10) -> list[datetime.date]:
    """指定した過去年数の日本の祝日を取得する。"""
    if year is None:
        year = datetime.datetime.now().year  # noqa: DTZ005

    dates = holidays.country_holidays("JP", years=range(year - n, year + 1))
    return sorted(dates.keys())


def with_date(df: DataFrame, year: int | None = None) -> DataFrame:
    """`Date`列を追加する。

    開示日が休日のとき、あるいは、開示時刻が15時以降の場合、Dateを開示日の翌営業日に設定する。
    """
    is_after_hours = pl.col("DisclosedTime").is_null() | (
        pl.col("DisclosedTime") > datetime.time(15, 0)
    )

    holidays = get_holidays(year=year)

    return df.select(
        pl.when(is_after_hours)
        .then(pl.col("DisclosedDate") + datetime.timedelta(days=1))
        .otherwise(pl.col("DisclosedDate"))
        .dt.add_business_days(0, holidays=holidays, roll="forward")
        .alias("Date"),
        pl.all(),
    )
